# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import pytest
import copy
import json
import re
import time
from datetime import datetime, timezone
from azure.core import MatchConditions
from azure.core.exceptions import (
    AzureError,
    ResourceModifiedError,
    ResourceNotFoundError,
    ResourceExistsError,
    HttpResponseError,
)
from azure.core.rest import HttpRequest
from azure.appconfiguration import (
    ResourceReadOnlyError,
    AzureAppConfigurationClient,
    ConfigurationSetting,
    ConfigurationSettingsFilter,
    FeatureFlagConfigurationSetting,
    SecretReferenceConfigurationSetting,
    FILTER_PERCENTAGE,
    FILTER_TARGETING,
    FILTER_TIME_WINDOW,
)
from testcase import AppConfigTestCase
from consts import (
    KEY,
    LABEL,
    TEST_VALUE,
    TEST_CONTENT_TYPE,
    LABEL_RESERVED_CHARS,
    PAGE_SIZE,
    KEY_UUID,
)
from uuid import uuid4
from preparers import app_config_decorator
from devtools_testutils import recorded_by_proxy, set_custom_default_matcher


class TestAppConfigurationClient(AppConfigTestCase):
    # method: add_configuration_setting
    @app_config_decorator
    @recorded_by_proxy
    def test_add_configuration_setting(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        test_config_setting = ConfigurationSetting(
            key=KEY + "_ADD",
            label=LABEL,
            value=TEST_VALUE,
            content_type=TEST_CONTENT_TYPE,
            tags={"tag1": "tag1", "tag2": "tag2"},
        )
        created_kv = client.add_configuration_setting(test_config_setting)
        assert (
            created_kv.label == test_config_setting.label
            and created_kv.value == test_config_setting.value
            and created_kv.content_type == test_config_setting.content_type
            and created_kv.tags == test_config_setting.tags
            and created_kv.etag != None
            and created_kv.etag != test_config_setting.etag
            and created_kv.last_modified != None
            and created_kv.read_only == False
        )

        # test add existing configuration setting
        with pytest.raises(ResourceExistsError):
            client.add_configuration_setting(
                ConfigurationSetting(
                    key=test_config_setting.key,
                    label=test_config_setting.label,
                )
            )
        client.delete_configuration_setting(key=created_kv.key, label=created_kv.label)

    # method: set_configuration_setting
    @app_config_decorator
    @recorded_by_proxy
    def test_set_existing_configuration_setting_label_etag(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        to_set_kv = self.create_config_setting()
        to_set_kv.value = to_set_kv.value + "a"
        to_set_kv.tags = {"a": "b", "c": "d"}
        set_kv = client.set_configuration_setting(to_set_kv)
        assert (
            to_set_kv.key == set_kv.key
            and to_set_kv.label == set_kv.label
            and to_set_kv.value == set_kv.value
            and to_set_kv.content_type == set_kv.content_type
            and to_set_kv.tags == set_kv.tags
            and to_set_kv.etag != set_kv.etag
        )
        client.delete_configuration_setting(key=to_set_kv.key, label=to_set_kv.label)

    @app_config_decorator
    @recorded_by_proxy
    def test_set_configuration_setting_wrong_etag(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        to_set_kv = self.create_config_setting()
        to_set_kv.value = to_set_kv.value + "a"
        to_set_kv.tags = {"a": "b", "c": "d"}
        to_set_kv.etag = "wrong etag"
        with pytest.raises(ResourceModifiedError):
            client.set_configuration_setting(to_set_kv, match_condition=MatchConditions.IfNotModified)

    # method: get_configuration_setting
    @app_config_decorator
    @recorded_by_proxy
    def test_get_configuration_setting_no_label(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        compare_kv = self.create_config_setting_no_label()
        self.add_for_test(client, compare_kv)
        fetched_kv = client.get_configuration_setting(compare_kv.key)
        assert (
            fetched_kv.key == compare_kv.key
            and fetched_kv.value == compare_kv.value
            and fetched_kv.content_type == compare_kv.content_type
            and fetched_kv.tags == compare_kv.tags
        )
        assert fetched_kv.label is None
        client.delete_configuration_setting(key=compare_kv.key, label=compare_kv.label)

    # method: get_configuration_setting
    @app_config_decorator
    @recorded_by_proxy
    def test_get_configuration_setting(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        compare_kv = self.create_config_setting()
        self.add_for_test(client, compare_kv)
        fetched_kv = client.get_configuration_setting(compare_kv.key, compare_kv.label)
        assert (
            fetched_kv.key == compare_kv.key
            and fetched_kv.value == compare_kv.value
            and fetched_kv.content_type == compare_kv.content_type
            and fetched_kv.tags == compare_kv.tags
            and fetched_kv.label == compare_kv.label
        )
        assert fetched_kv.label is not None
        client.delete_configuration_setting(key=compare_kv.key, label=compare_kv.label)

    @app_config_decorator
    @recorded_by_proxy
    def test_get_non_existing_configuration_setting(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        compare_kv = self.create_config_setting()
        with pytest.raises(ResourceNotFoundError):
            client.get_configuration_setting(compare_kv.key, compare_kv.label + "a")

    @app_config_decorator
    @recorded_by_proxy
    def test_get_configuration_setting_with_etag(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        compare_kv = self.create_config_setting()
        self.add_for_test(client, compare_kv)
        compare_kv = client.get_configuration_setting(compare_kv.key, compare_kv.label)

        # test get with wrong etag
        with pytest.raises(ResourceModifiedError):
            client.get_configuration_setting(
                compare_kv.key, compare_kv.label, etag="wrong etag", match_condition=MatchConditions.IfNotModified
            )
        # test get with correct etag
        with pytest.raises(ResourceNotFoundError):
            client.get_configuration_setting(compare_kv.key, etag=compare_kv.etag)
        client.get_configuration_setting(compare_kv.key, compare_kv.label, etag=compare_kv.etag)

        client.delete_configuration_setting(key=compare_kv.key, label=compare_kv.label)

    # method: delete_configuration_setting
    @app_config_decorator
    @recorded_by_proxy
    def test_delete_configuration_setting_with_key_no_label(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        to_delete_kv = self.create_config_setting_no_label()
        self.add_for_test(client, to_delete_kv)
        deleted_kv = client.delete_configuration_setting(key=to_delete_kv.key, label=to_delete_kv.label)
        assert deleted_kv is not None
        with pytest.raises(ResourceNotFoundError):
            client.get_configuration_setting(to_delete_kv.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_delete_configuration_setting_with_key_label(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        to_delete_kv = self.create_config_setting()
        self.add_for_test(client, to_delete_kv)
        deleted_kv = client.delete_configuration_setting(key=to_delete_kv.key, label=to_delete_kv.label)
        assert deleted_kv is not None
        with pytest.raises(ResourceNotFoundError):
            client.get_configuration_setting(to_delete_kv.key, label=to_delete_kv.label)

    @app_config_decorator
    @recorded_by_proxy
    def test_delete_not_existing_configuration_setting(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        deleted_kv = client.delete_configuration_setting("not_exist_" + KEY)
        assert deleted_kv is None

    @app_config_decorator
    @recorded_by_proxy
    def test_delete_configuration_setting_with_etag(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        to_delete_kv = self.create_config_setting_no_label()
        self.add_for_test(client, to_delete_kv)
        to_delete_kv = client.get_configuration_setting(to_delete_kv.key, to_delete_kv.label)

        # test delete with wrong etag
        with pytest.raises(ResourceModifiedError):
            client.delete_configuration_setting(
                to_delete_kv.key, etag="wrong etag", match_condition=MatchConditions.IfNotModified
            )
        # test delete with correct etag
        deleted_kv = client.delete_configuration_setting(to_delete_kv.key, etag=to_delete_kv.etag)
        assert deleted_kv is not None
        with pytest.raises(ResourceNotFoundError):
            client.get_configuration_setting(to_delete_kv.key)

    # method: list_configuration_settings
    @app_config_decorator
    @recorded_by_proxy
    def test_list_configuration_settings_key_label(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)
        items = list(self.client.list_configuration_settings(key_filter=KEY, label_filter=LABEL))
        assert len(items) == 1
        assert all(x.key == KEY and x.label == LABEL for x in items)

        with pytest.raises(TypeError) as ex:
            self.client.list_configuration_settings("MyKey1", key_filter="MyKey2")
        assert (
            str(ex.value)
            == "AzureAppConfigurationClient.list_configuration_settings() got multiple values for argument 'key_filter'"
        )
        with pytest.raises(TypeError) as ex:
            self.client.list_configuration_settings("MyKey", "MyLabel1", label_filter="MyLabel2")
        assert (
            str(ex.value)
            == "AzureAppConfigurationClient.list_configuration_settings() got multiple values for argument 'label_filter'"
        )
        with pytest.raises(TypeError) as ex:
            self.client.list_configuration_settings("None", key_filter="MyKey")
        assert (
            str(ex.value)
            == "AzureAppConfigurationClient.list_configuration_settings() got multiple values for argument 'key_filter'"
        )
        with pytest.raises(TypeError) as ex:
            self.client.list_configuration_settings("None", "None", label_filter="MyLabel")
        assert (
            str(ex.value)
            == "AzureAppConfigurationClient.list_configuration_settings() got multiple values for argument 'label_filter'"
        )

        self.tear_down()

    @app_config_decorator
    @recorded_by_proxy
    def test_list_configuration_settings_only_label(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)
        items = list(self.client.list_configuration_settings(label_filter=LABEL))
        assert len(items) == 1
        assert all(x.label == LABEL for x in items)
        self.tear_down()

    @app_config_decorator
    @recorded_by_proxy
    def test_list_configuration_settings_only_key(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)
        items = list(self.client.list_configuration_settings(key_filter=KEY))
        assert len(items) == 2
        assert all(x.key == KEY for x in items)
        self.tear_down()

    @app_config_decorator
    @recorded_by_proxy
    def test_list_configuration_settings_with_tags_filter(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)
        items = list(self.client.list_configuration_settings(tags_filter=["tag1=value1"]))
        assert len(items) == 1
        assert items[0].key == KEY
        assert items[0].label == LABEL
        self.tear_down()

    @app_config_decorator
    @recorded_by_proxy
    def test_list_configuration_settings_fields(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)
        items = list(
            self.client.list_configuration_settings(key_filter="*", label_filter=LABEL, fields=["key", "content_type"])
        )
        assert len(items) == 1
        assert all(x.key and not x.label and x.content_type for x in items)
        self.tear_down()

    @app_config_decorator
    @recorded_by_proxy
    def test_list_configuration_settings_reserved_chars(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        client = self.create_client(appconfiguration_connection_string)
        reserved_char_kv = ConfigurationSetting(key=KEY, label=LABEL_RESERVED_CHARS, value=TEST_VALUE)
        reserved_char_kv = client.add_configuration_setting(reserved_char_kv)
        escaped_label = re.sub(r"((?!^)\*(?!$)|\\|,)", r"\\\1", LABEL_RESERVED_CHARS)
        items = list(client.list_configuration_settings(label_filter=escaped_label))
        assert len(items) == 1
        assert all(x.label == LABEL_RESERVED_CHARS for x in items)
        client.delete_configuration_setting(reserved_char_kv.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_list_configuration_settings_contains(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)
        items = list(self.client.list_configuration_settings(label_filter=LABEL + "*"))
        assert len(items) == 1
        assert all(x.label == LABEL for x in items)
        self.tear_down()

    @app_config_decorator
    @recorded_by_proxy
    def test_list_configuration_settings_correct_etag(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        client = self.create_client(appconfiguration_connection_string)
        to_list_kv = self.create_config_setting()
        self.add_for_test(client, to_list_kv)
        to_list_kv = client.get_configuration_setting(to_list_kv.key, to_list_kv.label)
        custom_headers = {"If-Match": to_list_kv.etag}
        items = list(
            client.list_configuration_settings(
                key_filter=to_list_kv.key, label_filter=to_list_kv.label, headers=custom_headers
            )
        )
        assert len(items) == 1
        assert all(x.key == to_list_kv.key and x.label == to_list_kv.label for x in items)
        client.delete_configuration_setting(to_list_kv.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_list_configuration_settings_multi_pages(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        client = self.create_client(appconfiguration_connection_string)
        # create PAGE_SIZE+1 configuration settings to have at least two pages
        try:
            [
                client.add_configuration_setting(
                    ConfigurationSetting(
                        key="multi_" + str(i) + KEY_UUID,
                        label="multi_label_" + str(i),
                        value="multi value",
                    )
                )
                for i in range(PAGE_SIZE + 1)
            ]
        except ResourceExistsError:
            pass
        items = client.list_configuration_settings(key_filter="multi_*")
        assert len(list(items)) > PAGE_SIZE

        # Remove the configuration settings
        try:
            [
                client.delete_configuration_setting(key="multi_" + str(i) + KEY_UUID, label="multi_label_" + str(i))
                for i in range(PAGE_SIZE + 1)
            ]
        except AzureError:
            pass

    @app_config_decorator
    @recorded_by_proxy
    def test_list_configuration_settings_no_label(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)
        items = self.client.list_configuration_settings(label_filter="\0")
        assert len(list(items)) > 0
        self.tear_down()

    @app_config_decorator
    @recorded_by_proxy
    def test_list_configuration_settings_only_accepttime(self, appconfiguration_connection_string, **kwargs):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        recorded_variables = kwargs.pop("variables", {})
        recorded_variables.setdefault("timestamp", str(datetime.now(timezone.utc)))

        with self.create_client(appconfiguration_connection_string) as client:
            # Confirm all configuration settings are cleaned up
            current_config_settings = client.list_configuration_settings()
            if len(list(current_config_settings)) != 0:
                for config_setting in current_config_settings:
                    client.delete_configuration_setting(config_setting)

            revision = client.list_configuration_settings(accept_datetime=recorded_variables.get("timestamp"))
            assert len(list(revision)) >= 0

            accept_time = datetime(year=2000, month=4, day=1, hour=9, minute=30, second=45, tzinfo=timezone.utc)
            revision = client.list_configuration_settings(accept_datetime=accept_time)
            assert len(list(revision)) == 0

        return recorded_variables

    # method: list_revisions
    @app_config_decorator
    @recorded_by_proxy
    def test_list_revisions_key_label(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)
        to_list = self.create_config_setting()
        items = list(self.client.list_revisions(label_filter=to_list.label, key_filter=to_list.key))
        assert len(items) >= 2
        assert all(x.key == to_list.key and x.label == to_list.label for x in items)
        self.tear_down()

    @app_config_decorator
    @recorded_by_proxy
    def test_list_revisions_only_label(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)
        items = list(self.client.list_revisions(label_filter=LABEL))
        assert len(items) >= 1
        assert all(x.label == LABEL for x in items)
        self.tear_down()

    @app_config_decorator
    @recorded_by_proxy
    def test_list_revisions_key_no_label(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)
        items = list(self.client.list_revisions(key_filter=KEY))
        assert len(items) >= 1
        assert all(x.key == KEY for x in items)
        self.tear_down()

    @app_config_decorator
    @recorded_by_proxy
    def test_list_revisions_with_tags_filter(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)
        items = list(self.client.list_revisions(tags_filter=["tag1=value1"]))
        assert len(items) >= 1
        assert all(x.key == KEY for x in items)
        assert all(x.label == LABEL for x in items)
        self.tear_down()

    @app_config_decorator
    @recorded_by_proxy
    def test_list_revisions_fields(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)
        items = list(self.client.list_revisions(key_filter="*", label_filter=LABEL, fields=["key", "content_type"]))
        assert all(x.key and not x.label and x.content_type and not x.tags and not x.etag for x in items)
        self.tear_down()

    @app_config_decorator
    @recorded_by_proxy
    def test_list_revisions_correct_etag(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        to_list_kv = self.create_config_setting()
        self.add_for_test(client, to_list_kv)
        to_list_kv = client.get_configuration_setting(to_list_kv.key, to_list_kv.label)
        custom_headers = {"If-Match": to_list_kv.etag}
        items = list(
            client.list_revisions(key_filter=to_list_kv.key, label_filter=to_list_kv.label, headers=custom_headers)
        )
        assert len(items) >= 1
        assert all(x.key == to_list_kv.key and x.label == to_list_kv.label for x in items)

        client.delete_configuration_setting(to_list_kv.key)

    # method: set_read_only
    @app_config_decorator
    @recorded_by_proxy
    def test_set_read_only(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        to_set_kv = self.create_config_setting()
        self.add_for_test(client, to_set_kv)
        to_set_kv = client.get_configuration_setting(to_set_kv.key, to_set_kv.label)

        read_only_kv = client.set_read_only(to_set_kv)
        assert read_only_kv.read_only
        with pytest.raises(ResourceReadOnlyError):
            client.set_configuration_setting(read_only_kv)
        with pytest.raises(ResourceReadOnlyError):
            client.delete_configuration_setting(read_only_kv.key, read_only_kv.label)

        writable_kv = client.set_read_only(read_only_kv, False)
        assert not writable_kv.read_only
        client.set_configuration_setting(writable_kv)
        client.delete_configuration_setting(writable_kv.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_set_read_only_with_wrong_etag(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        to_set_kv = self.create_config_setting()
        self.add_for_test(client, to_set_kv)
        to_set_kv = client.get_configuration_setting(to_set_kv.key, to_set_kv.label)

        to_set_kv.etag = "wrong etag"
        with pytest.raises(ResourceModifiedError):
            client.set_read_only(to_set_kv, False, match_condition=MatchConditions.IfNotModified)

        client.delete_configuration_setting(to_set_kv)

    @app_config_decorator
    @recorded_by_proxy
    def test_sync_tokens_with_configuration_setting(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        sync_tokens = copy.deepcopy(client._sync_token_policy._sync_tokens)
        sync_token_header = self._order_dict(sync_tokens)
        sync_token_header = ",".join(str(x) for x in sync_token_header.values())

        new = ConfigurationSetting(
            key="KEY1",
            label=None,
            value="TEST_VALUE1",
            content_type=TEST_CONTENT_TYPE,
            tags={"tag1": "tag1", "tag2": "tag2"},
        )

        client.set_configuration_setting(new)
        sync_tokens2 = copy.deepcopy(client._sync_token_policy._sync_tokens)
        sync_token_header2 = self._order_dict(sync_tokens2)
        sync_token_header2 = ",".join(str(x) for x in sync_token_header2.values())
        assert sync_token_header != sync_token_header2

        new = ConfigurationSetting(
            key="KEY2",
            label=None,
            value="TEST_VALUE2",
            content_type=TEST_CONTENT_TYPE,
            tags={"tag1": "tag1", "tag2": "tag2"},
        )

        client.set_configuration_setting(new)
        sync_tokens3 = copy.deepcopy(client._sync_token_policy._sync_tokens)
        sync_token_header3 = self._order_dict(sync_tokens3)
        sync_token_header3 = ",".join(str(x) for x in sync_token_header3.values())
        assert sync_token_header2 != sync_token_header3

        client.delete_configuration_setting("KEY1")
        client.delete_configuration_setting("KEY2")

    @app_config_decorator
    @recorded_by_proxy
    def test_sync_tokens_with_feature_flag_configuration_setting(self, appconfiguration_connection_string):
        self.set_up(appconfiguration_connection_string)
        new = FeatureFlagConfigurationSetting(
            "custom",
            enabled=True,
            filters=[
                {
                    "name": "Microsoft.Percentage",
                    "parameters": {
                        "Value": 10,
                        "User": "user1",
                    },
                }
            ],
        )

        sync_tokens = copy.deepcopy(self.client._sync_token_policy._sync_tokens)
        keys = list(sync_tokens.keys())
        seq_num = sync_tokens[keys[0]].sequence_number
        self.client.set_configuration_setting(new)

        new = FeatureFlagConfigurationSetting(
            "time_window",
            enabled=True,
            filters=[
                {
                    "name": FILTER_TIME_WINDOW,
                    "parameters": {"Start": "Wed, 10 Mar 2021 05:00:00 GMT", "End": "Fri, 02 Apr 2021 04:00:00 GMT"},
                },
            ],
        )

        self.client.set_configuration_setting(new)
        sync_tokens2 = copy.deepcopy(self.client._sync_token_policy._sync_tokens)
        keys = list(sync_tokens2.keys())
        seq_num2 = sync_tokens2[keys[0]].sequence_number

        new = FeatureFlagConfigurationSetting(
            "newflag",
            enabled=True,
            filters=[
                {
                    "name": FILTER_TARGETING,
                    "parameters": {
                        "Audience": {"Users": ["abc", "def"], "Groups": ["ghi", "jkl"], "DefaultRolloutPercentage": 75}
                    },
                },
            ],
        )

        self.client.set_configuration_setting(new)
        sync_tokens3 = copy.deepcopy(self.client._sync_token_policy._sync_tokens)
        keys = list(sync_tokens3.keys())
        seq_num3 = sync_tokens3[keys[0]].sequence_number

        assert seq_num < seq_num2
        assert seq_num2 < seq_num3

        self.client.delete_configuration_setting(new.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_config_setting_feature_flag(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        feature_flag = FeatureFlagConfigurationSetting("test_feature", enabled=True)

        set_flag = client.set_configuration_setting(feature_flag)
        self._assert_same_keys(feature_flag, set_flag)
        set_flag_value = json.loads(set_flag.value)
        assert set_flag_value["id"] == "test_feature"
        assert set_flag_value["enabled"] == True
        assert set_flag_value["conditions"] != None

        set_flag.enabled = not set_flag.enabled
        changed_flag = client.set_configuration_setting(set_flag)
        assert changed_flag.enabled == False
        temp = json.loads(changed_flag.value)
        assert temp["id"] == set_flag_value["id"]
        assert temp["enabled"] == False
        assert temp["conditions"] == set_flag_value["conditions"]

        c = json.loads(changed_flag.value)
        c["enabled"] = True
        changed_flag.value = json.dumps(c)
        assert changed_flag.enabled == True
        temp = json.loads(changed_flag.value)
        assert temp["id"] == set_flag_value["id"]
        assert temp["enabled"] == True
        assert temp["conditions"] == set_flag_value["conditions"]

        changed_flag.value = json.dumps({})
        assert changed_flag.enabled == False
        temp = json.loads(changed_flag.value)
        assert temp["id"] == set_flag_value["id"]
        assert temp["enabled"] == False
        assert temp["conditions"] != None
        assert temp["conditions"]["client_filters"] == None

        set_flag.value = "bad_value"
        assert set_flag.enabled == False
        assert set_flag.filters == None
        assert set_flag.value == "bad_value"

        client.delete_configuration_setting(changed_flag.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_config_setting_secret_reference(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        secret_reference = SecretReferenceConfigurationSetting(
            "ConnectionString", "https://test-test.vault.azure.net/secrets/connectionString"
        )
        set_flag = client.set_configuration_setting(secret_reference)
        self._assert_same_keys(secret_reference, set_flag)

        set_flag.secret_id = "https://test-test.vault.azure.net/new_secrets/connectionString"
        updated_flag = client.set_configuration_setting(set_flag)
        self._assert_same_keys(set_flag, updated_flag)

        assert isinstance(updated_flag, SecretReferenceConfigurationSetting)
        new_uri = "https://aka.ms/azsdk"
        new_uri2 = "https://aka.ms/azsdk/python"
        updated_flag.secret_id = new_uri
        temp = json.loads(updated_flag.value)
        assert temp["uri"] == new_uri

        updated_flag.value = json.dumps({"uri": new_uri2})
        assert updated_flag.secret_id == new_uri2

        set_flag.value = "bad_value"
        assert set_flag.secret_id == None

        client.delete_configuration_setting(secret_reference.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_feature_filter_targeting(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        new = FeatureFlagConfigurationSetting(
            "newflag",
            enabled=True,
            filters=[
                {
                    "name": FILTER_TARGETING,
                    "parameters": {
                        "Audience": {"Users": ["abc", "def"], "Groups": ["ghi", "jkl"], "DefaultRolloutPercentage": 75}
                    },
                }
            ],
        )

        sent_config = client.set_configuration_setting(new)
        self._assert_same_keys(sent_config, new)

        assert isinstance(sent_config.filters[0], dict)
        assert len(sent_config.filters) == 1

        sent_config.filters[0]["parameters"]["Audience"]["DefaultRolloutPercentage"] = 80
        updated_sent_config = client.set_configuration_setting(sent_config)
        self._assert_same_keys(sent_config, updated_sent_config)

        updated_sent_config.filters.append(
            {
                "name": FILTER_TARGETING,
                "parameters": {
                    "Audience": {
                        "Users": ["abcd", "defg"],  # cspell:disable-line
                        "Groups": ["ghij", "jklm"],  # cspell:disable-line
                        "DefaultRolloutPercentage": 50,
                    }
                },
            }
        )
        updated_sent_config.filters.append(
            {
                "name": FILTER_TARGETING,
                "parameters": {
                    "Audience": {
                        "Users": ["abcde", "defgh"],  # cspell:disable-line
                        "Groups": ["ghijk", "jklmn"],  # cspell:disable-line
                        "DefaultRolloutPercentage": 100,
                    }
                },
            }
        )
        sent_config = client.set_configuration_setting(updated_sent_config)
        self._assert_same_keys(sent_config, updated_sent_config)
        assert len(sent_config.filters) == 3

        client.delete_configuration_setting(updated_sent_config.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_feature_filter_time_window(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        new = FeatureFlagConfigurationSetting(
            "time_window",
            enabled=True,
            filters=[
                {
                    "name": FILTER_TIME_WINDOW,
                    "parameters": {"Start": "Wed, 10 Mar 2021 05:00:00 GMT", "End": "Fri, 02 Apr 2021 04:00:00 GMT"},
                }
            ],
        )

        sent = client.set_configuration_setting(new)
        self._assert_same_keys(sent, new)

        sent.filters[0]["parameters"]["Start"] = "Thurs, 11 Mar 2021 05:00:00 GMT"
        new_sent = client.set_configuration_setting(sent)
        self._assert_same_keys(sent, new_sent)

        client.delete_configuration_setting(new_sent.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_feature_filter_custom(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        new = FeatureFlagConfigurationSetting(
            "custom", enabled=True, filters=[{"name": FILTER_PERCENTAGE, "parameters": {"Value": 10, "User": "user1"}}]
        )

        sent = client.set_configuration_setting(new)
        self._assert_same_keys(sent, new)

        sent.filters[0]["parameters"]["Value"] = 100
        new_sent = client.set_configuration_setting(sent)
        self._assert_same_keys(sent, new_sent)

        client.delete_configuration_setting(new_sent.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_feature_filter_multiple(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        new = FeatureFlagConfigurationSetting(
            "custom",
            enabled=True,
            filters=[
                {"name": FILTER_PERCENTAGE, "parameters": {"Value": 10}},
                {
                    "name": FILTER_TIME_WINDOW,
                    "parameters": {"Start": "Wed, 10 Mar 2021 05:00:00 GMT", "End": "Fri, 02 Apr 2021 04:00:00 GMT"},
                },
                {
                    "name": FILTER_TARGETING,
                    "parameters": {
                        "Audience": {
                            "Users": ["abcde", "defgh"],  # cspell:disable-line
                            "Groups": ["ghijk", "jklmn"],  # cspell:disable-line
                            "DefaultRolloutPercentage": 100,
                        }
                    },
                },
            ],
        )

        sent = client.set_configuration_setting(new)
        self._assert_same_keys(sent, new)

        sent.filters[0]["parameters"]["Value"] = 100
        sent.filters[1]["parameters"]["Start"] = "Wed, 10 Mar 2021 08:00:00 GMT"
        sent.filters[2]["parameters"]["Audience"]["DefaultRolloutPercentage"] = 100

        new_sent = client.set_configuration_setting(sent)
        self._assert_same_keys(sent, new_sent)

        assert new_sent.filters[0]["parameters"]["Value"] == 100
        assert new_sent.filters[1]["parameters"]["Start"] == "Wed, 10 Mar 2021 08:00:00 GMT"
        assert new_sent.filters[2]["parameters"]["Audience"]["DefaultRolloutPercentage"] == 100

        client.delete_configuration_setting(new_sent.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_feature_custom_fields(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        custom_fields = {
            "variants": [
                {"name": "Off", "configuration_value": "Off", "status_override": "Enabled"},
                {"name": "On", "configuration_value": "On", "status_override": "Disabled"},
            ],
            "allocation": {
                "percentile": [{"variant": "Off", "from": 0, "to": 100}],
                "user": [{"variant": "Off", "users": ["Adam"]}],
                "seed": "adfsasdfzsd",
                "default_when_enabled": "Off",
                "default_when_disabled": "Off",
            },
        }
        new = FeatureFlagConfigurationSetting(
            "Beta",
            description="Matt's variant FF",
            enabled=True,
        )
        new.value = json.dumps(custom_fields)
        sent = client.set_configuration_setting(new)
        self._assert_same_keys(sent, new)

        client.delete_configuration_setting(new.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_breaking_with_feature_flag_configuration_setting(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        new = FeatureFlagConfigurationSetting(
            "breaking1",
            enabled=True,
            filters=[
                {
                    "name": FILTER_TIME_WINDOW,
                    "parameters": {
                        "Start": "bababooey, 31 Mar 2021 25:00:00 GMT",  # cspell:disable-line
                        "End": "Fri, 02 Apr 2021 04:00:00 GMT",
                    },
                },
            ],
        )
        client.set_configuration_setting(new)
        client.get_configuration_setting(new.key)
        client.delete_configuration_setting(new.key)

        new = FeatureFlagConfigurationSetting(
            "breaking2",
            enabled=True,
            filters=[
                {
                    "name": FILTER_TIME_WINDOW,
                    "parameters": {
                        "Start": "bababooey, 31 Mar 2021 25:00:00 GMT",  # cspell:disable-line
                        "End": "not even trying to be a date",
                    },
                },
            ],
        )
        client.set_configuration_setting(new)
        client.get_configuration_setting(new.key)
        client.delete_configuration_setting(new.key)

        # This will show up as a Custom filter
        new = FeatureFlagConfigurationSetting(
            "breaking3",
            enabled=True,
            filters=[
                {
                    "name": FILTER_TIME_WINDOW,
                    "parameters": {
                        "Start": "bababooey, 31 Mar 2021 25:00:00 GMT",  # cspell:disable-line
                        "End": "not even trying to be a date",
                    },
                },
            ],
        )
        client.set_configuration_setting(new)
        client.get_configuration_setting(new.key)
        client.delete_configuration_setting(new.key)

        new = FeatureFlagConfigurationSetting(
            "breaking4",
            enabled=True,
            filters=[
                {"name": FILTER_TIME_WINDOW, "parameters": "stringystring"},
            ],
        )
        client.set_configuration_setting(new)
        client.get_configuration_setting(new.key)
        client.delete_configuration_setting(new.key)

        new = FeatureFlagConfigurationSetting(
            "breaking5",
            enabled=True,
            filters=[{"name": FILTER_TARGETING, "parameters": {"Audience": {"Users": "123"}}}],
        )
        client.set_configuration_setting(new)
        client.get_configuration_setting(new.key)
        client.delete_configuration_setting(new.key)

        new = FeatureFlagConfigurationSetting(
            "breaking6", enabled=True, filters=[{"name": FILTER_TARGETING, "parameters": "invalidformat"}]
        )
        client.set_configuration_setting(new)
        client.get_configuration_setting(new.key)
        client.delete_configuration_setting(new.key)

        new = FeatureFlagConfigurationSetting("breaking7", enabled=True, filters=[{"abc": "def"}])
        client.set_configuration_setting(new)
        client.get_configuration_setting(new.key)
        client.delete_configuration_setting(new.key)

        new = FeatureFlagConfigurationSetting("breaking8", enabled=True, filters=[{"abc": "def"}])
        new.feature_flag_content_type = "fakeyfakey"  # cspell:disable-line
        client.set_configuration_setting(new)
        client.get_configuration_setting(new.key)
        client.delete_configuration_setting(new.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_breaking_with_secret_reference_configuration_setting(self, appconfiguration_connection_string):
        client = self.create_client(appconfiguration_connection_string)
        new = SecretReferenceConfigurationSetting("aref", "notaurl")  # cspell:disable-line
        client.set_configuration_setting(new)
        client.get_configuration_setting(new.key)

        client.delete_configuration_setting(new.key)

        new = SecretReferenceConfigurationSetting("aref1", "notaurl")  # cspell:disable-line
        new.content_type = "fkaeyjfdkal;"  # cspell:disable-line
        client.set_configuration_setting(new)
        client.get_configuration_setting(new.key)

        client.delete_configuration_setting(new.key)

    @app_config_decorator
    @recorded_by_proxy
    def test_create_snapshot(self, appconfiguration_connection_string, **kwargs):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)

        variables = kwargs.pop("variables", {})
        dynamic_snapshot_name_postfix = variables.setdefault("dynamic_snapshot_name_postfix", str(int(time.time())))

        snapshot_name = f"{self.get_resource_name('snapshot')}_{dynamic_snapshot_name_postfix}"

        filters = [ConfigurationSettingsFilter(key=KEY, label=LABEL)]
        response = self.client.begin_create_snapshot(name=snapshot_name, filters=filters, retention_period=3600)
        created_snapshot = response.result()
        assert created_snapshot.name == snapshot_name
        assert created_snapshot.status == "ready"
        assert created_snapshot.retention_period == 3600
        assert len(created_snapshot.filters) == 1
        assert created_snapshot.filters[0].key == KEY
        assert created_snapshot.filters[0].label == LABEL

        received_snapshot = self.client.get_snapshot(name=snapshot_name)
        self._assert_snapshots(received_snapshot, created_snapshot)

        self.tear_down()
        return variables

    @app_config_decorator
    @recorded_by_proxy
    def test_update_snapshot_status(self, appconfiguration_connection_string, **kwargs):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)

        variables = kwargs.pop("variables", {})
        dynamic_snapshot_name_postfix = variables.setdefault("dynamic_snapshot_name_postfix", str(int(time.time())))

        snapshot_name = f"{self.get_resource_name('snapshot')}_{dynamic_snapshot_name_postfix}"

        filters = [ConfigurationSettingsFilter(key=KEY, label=LABEL)]
        response = self.client.begin_create_snapshot(name=snapshot_name, filters=filters, retention_period=3600)
        created_snapshot = response.result()
        assert created_snapshot.status == "ready"

        archived_snapshot = self.client.archive_snapshot(name=snapshot_name)
        assert archived_snapshot.status == "archived"

        recovered_snapshot = self.client.recover_snapshot(name=snapshot_name)
        assert recovered_snapshot.status == "ready"

        self.tear_down()
        return variables

    @app_config_decorator
    @recorded_by_proxy
    def test_update_snapshot_status_with_etag(self, appconfiguration_connection_string, **kwargs):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)

        variables = kwargs.pop("variables", {})
        dynamic_snapshot_name_postfix = variables.setdefault("dynamic_snapshot_name_postfix", str(int(time.time())))

        snapshot_name = f"{self.get_resource_name('snapshot')}_{dynamic_snapshot_name_postfix}"

        filters = [ConfigurationSettingsFilter(key=KEY, label=LABEL)]
        response = self.client.begin_create_snapshot(name=snapshot_name, filters=filters, retention_period=3600)
        created_snapshot = response.result()

        # test update with wrong etag
        with pytest.raises(ResourceModifiedError):
            self.client.archive_snapshot(
                name=snapshot_name, etag="wrong etag", match_condition=MatchConditions.IfNotModified
            )
        # test update with correct etag
        archived_snapshot = self.client.archive_snapshot(name=snapshot_name, etag=created_snapshot.etag)
        assert archived_snapshot.status == "archived"

        self.tear_down()
        return variables

    @app_config_decorator
    @recorded_by_proxy
    def test_list_snapshots(self, appconfiguration_connection_string, **kwargs):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)

        # Only list "ready" snapshots to avoid counting archived snapshots that may expire during test runs
        result = self.client.list_snapshots(status=["ready"])
        initial_snapshots = len(list(result))

        variables = kwargs.pop("variables", {})
        dynamic_snapshot_name_postfix = variables.setdefault("dynamic_snapshot_name_postfix", str(int(time.time())))

        snapshot_name1 = f"{self.get_resource_name('snapshot1')}_{dynamic_snapshot_name_postfix}"
        snapshot_name2 = f"{self.get_resource_name('snapshot2')}_{dynamic_snapshot_name_postfix}"

        filters1 = [ConfigurationSettingsFilter(key=KEY)]
        response1 = self.client.begin_create_snapshot(name=snapshot_name1, filters=filters1, retention_period=3600)
        created_snapshot1 = response1.result()
        assert created_snapshot1.status == "ready"
        filters2 = [ConfigurationSettingsFilter(key=KEY, label=LABEL)]
        response2 = self.client.begin_create_snapshot(name=snapshot_name2, filters=filters2, retention_period=3600)
        created_snapshot2 = response2.result()
        assert created_snapshot2.status == "ready"

        result = self.client.list_snapshots(status=["ready"])
        assert len(list(result)) == initial_snapshots + 2

        self.tear_down()
        return variables

    @app_config_decorator
    @recorded_by_proxy
    def test_list_snapshot_configuration_settings(self, appconfiguration_connection_string, **kwargs):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        self.set_up(appconfiguration_connection_string)

        variables = kwargs.pop("variables", {})
        dynamic_snapshot_name_postfix = variables.setdefault("dynamic_snapshot_name_postfix", str(int(time.time())))

        snapshot_name1 = f"{self.get_resource_name('snapshot1')}_{dynamic_snapshot_name_postfix}"

        filters = [ConfigurationSettingsFilter(key=KEY, label=LABEL)]
        response = self.client.begin_create_snapshot(name=snapshot_name1, filters=filters, retention_period=3600)
        created_snapshot = response.result()
        assert created_snapshot.status == "ready"

        items = self.client.list_configuration_settings(snapshot_name=snapshot_name1)
        assert len(list(items)) == 1

        snapshot_name2 = f"{self.get_resource_name('snapshot2')}_{dynamic_snapshot_name_postfix}"

        filters = [ConfigurationSettingsFilter(key=KEY, label=LABEL, tags=["tag1=invalid"])]
        response = self.client.begin_create_snapshot(name=snapshot_name2, filters=filters, retention_period=3600)
        created_snapshot = response.result()
        assert created_snapshot.status == "ready"

        items = self.client.list_configuration_settings(snapshot_name=snapshot_name2)
        assert len(list(items)) == 0

        self.tear_down()
        return variables

    @app_config_decorator
    @recorded_by_proxy
    def test_monitor_configuration_settings_by_page_etag(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        with AzureAppConfigurationClient.from_connection_string(appconfiguration_connection_string) as client:
            self.client = client
            # prepare 200 configuration settings
            for i in range(200):
                self.client.set_configuration_setting(
                    ConfigurationSetting(
                        key=f"sample_key_{str(i)}",
                        label=f"sample_label_{str(i)}",
                    )
                )
            # there will have 2 pages while listing, there are 100 configuration settings per page.

            # get page etags
            match_conditions = []
            items = self.client.list_configuration_settings(key_filter="sample_key_*", label_filter="sample_label_*")
            iterator = items.by_page()
            for _ in iterator:
                etag = iterator.etag
                match_conditions.append(etag)

            # monitor page updates without changes - only changed pages will be yielded
            items = self.client.list_configuration_settings(key_filter="sample_key_*", label_filter="sample_label_*")
            iterator = items.by_page(match_conditions=match_conditions)
            changed_pages = list(iterator)

            # No pages should be yielded since nothing changed
            assert len(changed_pages) == 0

            # do some changes
            self.client.set_configuration_setting(
                ConfigurationSetting(
                    key="sample_key_201",
                    label="sample_label_202",
                )
            )
            # now we have three pages, 100 settings in first two pages and 1 setting in the last page

            # get page etags after updates
            new_match_conditions = []
            items = self.client.list_configuration_settings(key_filter="sample_key_*", label_filter="sample_label_*")
            iterator = items.by_page()
            for _ in iterator:
                etag = iterator.etag
                new_match_conditions.append(etag)

            assert match_conditions[0] == new_match_conditions[0]
            assert match_conditions[1] != new_match_conditions[1]
            assert match_conditions[2] != new_match_conditions[2]
            assert len(new_match_conditions) == 3

            # monitor pages after updates - only changed pages will be yielded
            items = self.client.list_configuration_settings(key_filter="sample_key_*", label_filter="sample_label_*")
            iterator = items.by_page(match_conditions=new_match_conditions)
            changed_pages = list(iterator)
            # Should yield 0 pages
            assert len(changed_pages) == 0

            # clean up
            self.tear_down()

    @app_config_decorator
    @recorded_by_proxy
    def test_list_labels(self, appconfiguration_connection_string):
        self.set_up(appconfiguration_connection_string)

        rep = self.client.list_labels()
        assert len(list(rep)) >= 2

        rep = self.client.list_labels(name="test*")
        assert len(list(rep)) == 1

        rep = self.client.list_labels(name="test'@*$!%")
        with pytest.raises(HttpResponseError) as error:
            rep.next()
        assert error.value.status_code == 400
        assert error.value.message == "Operation returned an invalid status 'Bad Request'"
        assert (
            '"title":"Invalid request parameter \'label\'","name":"label","detail":"label(6): Invalid character"'
            in str(error.value)
        )

        self.tear_down()
        rep = self.client.list_labels()
        assert len(list(rep)) == 0


class TestAppConfigurationClientUnitTest:
    def test_type_error(self):
        with pytest.raises(TypeError):
            _ = FeatureFlagConfigurationSetting("blah", value="blah")
        with pytest.raises(TypeError):
            _ = SecretReferenceConfigurationSetting("blah", value="blah")

    def test_mock_policies(self):
        from azure.core.pipeline.transport import HttpResponse, HttpTransport
        from azure.core.pipeline import PipelineRequest, PipelineResponse
        from consts import APPCONFIGURATION_CONNECTION_STRING

        class MockTransport(HttpTransport):
            def __init__(self):
                self.auth_headers = []

            def __exit__(self, exc_type, exc_val, exc_tb):
                pass

            def close(self):
                pass

            def open(self):
                pass

            def send(self, request: PipelineRequest, **kwargs) -> PipelineResponse:
                assert request.headers["Authorization"] != self.auth_headers
                self.auth_headers.append(request.headers["Authorization"])
                response = HttpResponse(request, None)
                response.status_code = 429
                return response

        def new_method(self, request):
            request.http_request.headers["Authorization"] = str(uuid4())

        from azure.appconfiguration._azure_appconfiguration_requests import AppConfigRequestsCredentialsPolicy

        # Store the method to restore later
        temp = AppConfigRequestsCredentialsPolicy._signed_request
        AppConfigRequestsCredentialsPolicy._signed_request = new_method

        client = AzureAppConfigurationClient.from_connection_string(
            APPCONFIGURATION_CONNECTION_STRING, transport=MockTransport()
        )
        client.list_configuration_settings()

        # Reset the actual method
        AppConfigRequestsCredentialsPolicy._signed_request = temp



# coding: utf-8

# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""
FILE: list_labels_sample.py

DESCRIPTION:
    This sample demos how to list labels synchronously.

USAGE: python list_labels_sample.py

    Set the environment variables with your own values before running the sample:
    1) APPCONFIGURATION_CONNECTION_STRING: Connection String used to access the Azure App Configuration.
"""
import os
from azure.appconfiguration import AzureAppConfigurationClient, ConfigurationSetting


def main():
    CONNECTION_STRING = os.environ["APPCONFIGURATION_CONNECTION_STRING"]

    # Create an app config client
    client = AzureAppConfigurationClient.from_connection_string(CONNECTION_STRING)

    config_setting1 = ConfigurationSetting(
        key="MyKey1", label="my label1", content_type="my content type", tags={"my tag1": "my tag1 value"}
    )
    config_setting2 = ConfigurationSetting(
        key="MyKey2", label="my label2", content_type="my content type", tags={"my tag2": "my tag2 value"}
    )
    client.set_configuration_setting(config_setting1)
    client.set_configuration_setting(config_setting2)

    # [START list_labels]
    print("List all labels in resource")
    config_settings = client.list_labels()
    for config_setting in config_settings:
        print(config_setting)

    print("List labels by exact match")
    config_settings = client.list_labels(name="my label1")
    for config_setting in config_settings:
        print(config_setting)

    print("List labels by wildcard")
    config_settings = client.list_labels(name="my label*")
    for config_setting in config_settings:
        print(config_setting)
    # [END list_labels]

    client.delete_configuration_setting(key="MyKey1", label="my label1")
    client.delete_configuration_setting(key="MyKey2", label="my label2")


if __name__ == "__main__":
    main()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import pytest
from azure.core.exceptions import ClientAuthenticationError
from azure.appconfiguration._audience_error_handling_policy import (
    AudienceErrorHandlingPolicy,
    INCORRECT_AUDIENCE_ERROR_MESSAGE,
)
from asynctestcase import AsyncAppConfigTestCase
from async_preparers import app_config_aad_decorator_async
from devtools_testutils.aio import recorded_by_proxy_async

# cspell:disable-next-line
CORRECT_AUDIENCE = "https://azconfig.io"
# cspell:disable-next-line
VALID_ENDPOINT_SUFFIX = ".azconfig.io"
# cspell:disable-next-line
INVALID_ENDPOINT_SUFFIX = ".azconfig.sovcloud-api.fr"
# cspell:disable-next-line
INCORRECT_AUDIENCE = "https://login.sovcloud-identity2.fr"


class TestAudienceErrorHandlingLiveAsync(AsyncAppConfigTestCase):
    """Async live and recorded tests for audience error handling with the Azure App Configuration client."""

    @app_config_aad_decorator_async
    @recorded_by_proxy_async
    async def test_async_client_has_audience_policy_with_no_audience(self, appconfiguration_endpoint_string):
        """Test that async client created without audience has policy with has_audience=False."""
        # Create client without audience
        client = self.create_aad_client(appconfiguration_endpoint_string)

        # Check that audience error handling policy is in the pipeline
        policies = client._impl._client._pipeline._impl_policies
        audience_policy = None
        for policy in policies:
            # Policies are wrapped in _SansIOHTTPPolicyRunner, so we need to access the underlying policy
            underlying_policy = getattr(policy, "_policy", policy)
            if isinstance(underlying_policy, AudienceErrorHandlingPolicy):
                audience_policy = underlying_policy
                break

        assert audience_policy is not None, "AudienceErrorHandlingPolicy should be in pipeline"
        assert audience_policy.has_audience is False, "has_audience should be False when no audience provided"

        await client.close()

    @app_config_aad_decorator_async
    @recorded_by_proxy_async
    async def test_async_client_has_audience_policy_with_audience(self, appconfiguration_endpoint_string):
        """Test that async client created with audience has policy with has_audience=True."""
        # Create client with audience
        client = self.create_aad_client(appconfiguration_endpoint_string, audience=CORRECT_AUDIENCE)

        # Check that audience error handling policy is in the pipeline
        policies = client._impl._client._pipeline._impl_policies
        audience_policy = None
        for policy in policies:
            # Policies are wrapped in _SansIOHTTPPolicyRunner, so we need to access the underlying policy
            underlying_policy = getattr(policy, "_policy", policy)
            if isinstance(underlying_policy, AudienceErrorHandlingPolicy):
                audience_policy = underlying_policy
                break

        assert audience_policy is not None, "AudienceErrorHandlingPolicy should be in pipeline"
        assert audience_policy.has_audience is True, "has_audience should be True when audience provided"

        await client.close()



# coding=utf-8
# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# Code generated by Microsoft (R) Python Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is regenerated.
# --------------------------------------------------------------------------

from enum import Enum
from azure.core import CaseInsensitiveEnumMeta


class ConfigurationSettingFields(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Key-value fields."""

    KEY = "key"
    """Key field."""
    LABEL = "label"
    """Label field."""
    CONTENT_TYPE = "content_type"
    """Content type field."""
    VALUE = "value"
    """Value field."""
    LAST_MODIFIED = "last_modified"
    """Last modified field."""
    TAGS = "tags"
    """Tags field."""
    LOCKED = "locked"
    """Locked field."""
    ETAG = "etag"
    """Etag field."""


class LabelFields(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Label fields."""

    NAME = "name"
    """Name field."""


class OperationState(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Enum describing allowed operation states."""

    NOT_STARTED = "NotStarted"
    """The operation has not started."""
    RUNNING = "Running"
    """The operation is in progress."""
    SUCCEEDED = "Succeeded"
    """The operation has completed successfully."""
    FAILED = "Failed"
    """The operation has failed."""
    CANCELED = "Canceled"
    """The operation has been canceled by the user."""


class SnapshotComposition(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Composition types."""

    KEY = "key"
    """The 'key' composition type."""
    KEY_LABEL = "key_label"
    """The 'key_label' composition type."""


class SnapshotFields(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Snapshot fields."""

    NAME = "name"
    """Name field."""
    STATUS = "status"
    """Status field."""
    FILTERS = "filters"
    """Filters field."""
    COMPOSITION_TYPE = "composition_type"
    """Composition type field."""
    CREATED = "created"
    """Created field."""
    EXPIRES = "expires"
    """Expires field."""
    RETENTION_PERIOD = "retention_period"
    """Retention period field."""
    SIZE = "size"
    """Size field."""
    ITEMS_COUNT = "items_count"
    """Items count field."""
    TAGS = "tags"
    """Tags field."""
    ETAG = "etag"
    """Etag field."""


class SnapshotStatus(str, Enum, metaclass=CaseInsensitiveEnumMeta):
    """Snapshot status."""

    PROVISIONING = "provisioning"
    """Provisioning"""
    READY = "ready"
    """Ready"""
    ARCHIVED = "archived"
    """Archived"""
    FAILED = "failed"
    """Failed"""



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import base64
from types import SimpleNamespace
from azure.core.credentials import AzureKeyCredential
from azure.core.pipeline.transport import HttpRequest
from azure.core.pipeline import PipelineRequest
from azure.appconfiguration._azure_appconfiguration_requests import (
    AppConfigRequestsCredentialsPolicy,
)
from azure.appconfiguration._utils import parse_connection_string


def test_parse_connection_string_returns_http_endpoint():
    # An http (not https) endpoint to ensure scheme is preserved
    conn_str = "Endpoint=http://localhost:8483;Id=test-id;Secret=test-secret"
    endpoint, id_, secret = parse_connection_string(conn_str)
    assert endpoint == "http://localhost:8483"
    assert id_ == "test-id"
    assert secret == "test-secret"


def test_signed_request_with_http_endpoint():
    endpoint = "http://localhost:8483"
    request = HttpRequest("GET", endpoint + "/kv/foo?label=env")
    # Create pipeline request with placeholder context, then replace with a namespace that has a transport attr.
    pipeline_request = PipelineRequest(request, {})
    pipeline_request.context = SimpleNamespace(transport=object())

    key = base64.b64encode(b"secret").decode("utf-8")
    credential = AzureKeyCredential(key)
    policy = AppConfigRequestsCredentialsPolicy(credential, endpoint, "test-id")

    policy._signed_request(pipeline_request)  # pylint: disable=protected-access

    headers = pipeline_request.http_request.headers

    assert pipeline_request.http_request.url.startswith("http://")
    assert "x-ms-date" in headers
    assert "x-ms-content-sha256" in headers
    auth = headers.get("Authorization")
    assert auth is not None
    assert auth.startswith("HMAC-SHA256 ")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


from azure.core.pipeline.transport import HttpRequest
from azure.core.pipeline import PipelineRequest
from azure.appconfiguration._query_param_policy import QueryParamPolicy
import pytest

TEST_URL = "https://example.com"


def test_query_parameters_are_sorted_alphabetically():
    """Test that query parameters are sorted alphabetically."""
    original_url = "?zebra=value1&alpha=value2&beta=value3"
    expected_url = "?alpha=value2&beta=value3&zebra=value1"

    run_query_param_policy_test(original_url, expected_url)


def test_query_parameter_keys_are_converted_to_lowercase():
    """Test that query parameter keys are converted to lowercase."""
    original_url = "?SELECT=field1&FILTER=condition&orderBy=field2"
    expected_url = "?filter=condition&orderby=field2&select=field1"

    run_query_param_policy_test(original_url, expected_url)


def test_query_parameter_o_data_parameters():
    """Test that OData query parameters are handled correctly."""
    original_url = "?$Select=name%2Cvalue&$Filter=startsWith%28key%2C%27test%27%29&api-version=1.0"
    expected_url = "?%24filter=startsWith%28key%2C%27test%27%29&%24select=name%2Cvalue&api-version=1.0"

    run_query_param_policy_test(original_url, expected_url)


def test_query_parameters_with_multiple_values():
    """Test that query parameters with multiple values are handled correctly."""
    original_url = "?key=value1&key=value2&alpha=test"
    expected_url = "?alpha=test&key=value1&key=value2"

    run_query_param_policy_test(original_url, expected_url)


def test_query_parameters_with_special_characters():
    """Test that query parameters with special characters are preserved."""
    original_url = "?filter=name%20eq%20%27test%27&select=*"
    expected_url = "?filter=name%20eq%20%27test%27&select=%2A"

    request = HttpRequest("GET", original_url)
    pipeline_request = PipelineRequest(request, None)

    query_param_policy = QueryParamPolicy()

    # Create a mock next policy
    class MockNext:
        def __init__(self):
            self.captured_url = None

        def send(self, request):
            self.captured_url = request.http_request.url
            return None

    mock_next = MockNext()
    query_param_policy.next = mock_next

    query_param_policy.send(pipeline_request)

    # Note: URL encoding may change the format slightly, but the query params should still be sorted
    assert mock_next.captured_url is not None, "No URL was captured by the mock policy."
    assert "filter=" in mock_next.captured_url
    assert "select=" in mock_next.captured_url
    # The params should be in alphabetical order
    filter_pos = mock_next.captured_url.index("filter=")
    select_pos = mock_next.captured_url.index("select=")
    assert filter_pos < select_pos, "Parameters should be sorted alphabetically"


def test_no_query_parameters():
    """Test that URLs without query parameters are not modified."""
    original_url = ""
    expected_url = ""

    run_query_param_policy_test(original_url, expected_url)


def test_empty_query_parameter_values():
    """Test that empty query parameter values are preserved."""
    original_url = "?zebra=&alpha=value&beta="
    expected_url = "?alpha=value&beta=&zebra="

    run_query_param_policy_test(original_url, expected_url)


def test_query_parameter_with_key_only():
    """Test that query parameters with only a key and no value are preserved and sorted."""
    original_url = "?zebra&alpha=value&beta"
    expected_url = "?alpha=value&beta=&zebra="

    run_query_param_policy_test(original_url, expected_url)


def test_query_parameter_with_empty_key_is_first():
    """Test that query parameters with empty keys are first."""
    original_url = "?alpha=value2&=value1"
    expected_url = "?alpha=value2"

    run_query_param_policy_test(original_url, expected_url)


@pytest.mark.parametrize(
    "original_url,expected_url",
    [
        ("?key=%20value%20&alpha=%20%20", "?alpha=%20%20&key=%20value%20"),
        ("?key=hello%20world&alpha=foo%20bar", "?alpha=foo%20bar&key=hello%20world"),
    ],
)
def test_query_param_policy_whitespace(original_url, expected_url):
    run_query_param_policy_test(original_url, expected_url)


# Unicode and encoded values
@pytest.mark.parametrize(
    "original_url,expected_url",
    [
        ("?key=%E2%9C%93&alpha=%C3%A9", "?alpha=%C3%A9&key=%E2%9C%93"),
        ("?key=val1&key=\u2713", "?key=val1&key=%E2%9C%93"),
        ("?key=val1&key=é", "?key=val1&key=%C3%A9"),
    ],
)
def test_query_param_policy_unicode(original_url, expected_url):
    run_query_param_policy_test(original_url, expected_url)


# Multiple values for same key
@pytest.mark.parametrize(
    "original_url,expected_url",
    [
        ("?key=val1&key=val2&key=val3", "?key=val1&key=val2&key=val3"),
    ],
)
def test_query_param_policy_multiple_values(original_url, expected_url):
    run_query_param_policy_test(original_url, expected_url)


def test_query_parameter_values_are_preserved():
    """Test that query parameter values are preserved correctly."""
    original_url = "?key1=Value%20With%20Spaces&key2=SimpleValue&key3="
    expected_url = "?key1=Value%20With%20Spaces&key2=SimpleValue&key3="

    run_query_param_policy_test(original_url, expected_url)


def test_query_parameters_with_same_key_are_preserved():
    """Test that query parameters with the same key are preserved in order."""
    original_url = "?key=val1&key=val2&key=val3"
    expected_url = "?key=val1&key=val2&key=val3"

    run_query_param_policy_test(original_url, expected_url)


# Empty and malformed cases
@pytest.mark.parametrize(
    "original_url,expected_url",
    [
        ("?key=val1&&key=val2", "?key=val1&key=val2"),
        ("?key=val1&=val2", "?key=val1"),
        ("?key=val1&key=", "?key=val1&key="),
        ("?key=val1&key", "?key=val1&key="),
        ("?key=val1&key= ", "?key=val1&key=%20"),
        ("?key=val1&key=%20", "?key=val1&key=%20"),
        ("?key=val1&key=%E2%9C%93", "?key=val1&key=%E2%9C%93"),
    ],
)
def test_query_param_policy_empty_and_malformed(original_url, expected_url):
    run_query_param_policy_test(original_url, expected_url)


def test_comprehensive_query_parameter_normalization():
    original_url = (
        "?$TOP=10&API-Version=2023-10-01&$select=key,value&label=prod&$filter=startsWith(key,'app')&maxItems=100"
    )
    expected_url = "?%24filter=startsWith%28key%2C%27app%27%29&%24select=key%2Cvalue&%24top=10&api-version=2023-10-01&label=prod&maxitems=100"

    run_query_param_policy_test(original_url, expected_url)


def test_multiple_tags_parameters():
    """Test that multiple tags parameters are handled correctly."""
    # cspell thinks the url encoding is part of the word
    original_url = (
        "?api-version=2023-11-01&key=*&label=dev&tags=environment%3Ddev&tags=team%3Dfrontend"  # cspell:disable-line
    )
    expected_url = (
        "?api-version=2023-11-01&key=%2A&label=dev&tags=environment%3Ddev&tags=team%3Dfrontend"  # cspell:disable-line
    )

    run_query_param_policy_test(original_url, expected_url)


def test_tags_parameters_with_complex_values():
    """Test that tags parameters with complex values are handled correctly."""
    original_url = "?tags=environment%3Dproduction&tags=team%3Dbackend&api-version=2023-11-01"  # cspell:disable-line
    expected_url = "?api-version=2023-11-01&tags=environment%3Dproduction&tags=team%3Dbackend"  # cspell:disable-line

    run_query_param_policy_test(original_url, expected_url)


def test_tags_parameters_mixed_with_other_parameters():
    """Test that tags parameters mixed with other parameters are handled correctly."""
    original_url = "?$select=key,value&tags=feature%3Dauth&label=*&api-version=2023-11-01&$filter=startsWith(key,'app')&tags=env%3Dtest"  # cspell:disable-line
    expected_url = "?%24filter=startsWith%28key%2C%27app%27%29&%24select=key%2Cvalue&api-version=2023-11-01&label=%2A&tags=feature%3Dauth&tags=env%3Dtest"  # cspell:disable-line

    run_query_param_policy_test(original_url, expected_url)


def test_tags_parameters_with_special_characters():
    """Test that tags parameters with special characters are handled correctly."""
    original_url = "?TAGS=Priority%3DHigh&api-version=2023-11-01&Tags=Status%3DActive"
    expected_url = "?api-version=2023-11-01&tags=Priority%3DHigh&tags=Status%3DActive"

    run_query_param_policy_test(original_url, expected_url)


def test_key_and_label_filters_with_ampersand_character():
    """Test that key and label filters with ampersand characters are handled correctly."""
    original_url = "?key=app%26config&label=prod%26test&api-version=2023-11-01"
    expected_url = "?api-version=2023-11-01&key=app%26config&label=prod%26test"

    run_query_param_policy_test(original_url, expected_url)


def test_key_and_label_filters_with_space_character():
    """Test that key and label filters with space characters are handled correctly."""
    original_url = "?key=app%20config&label=dev%20environment&api-version=2023-11-01"
    expected_url = "?api-version=2023-11-01&key=app%20config&label=dev%20environment"

    run_query_param_policy_test(original_url, expected_url)


def test_key_and_label_filters_with_hash_character():
    """Test that key and label filters with hash characters are handled correctly."""
    original_url = "?key=app%23config&label=version%23v1&api-version=2023-11-01"
    expected_url = "?api-version=2023-11-01&key=app%23config&label=version%23v1"

    run_query_param_policy_test(original_url, expected_url)


def test_key_and_label_filters_with_mixed_special_characters():
    """Test that key and label filters with mixed special characters are handled correctly."""
    original_url = "?key=app%26config%20test%23v1&label=prod%20%26%20test%23env&api-version=2023-11-01"
    expected_url = "?api-version=2023-11-01&key=app%26config%20test%23v1&label=prod%20%26%20test%23env"

    run_query_param_policy_test(original_url, expected_url)


def run_query_param_policy_test(original_url, expected_url):
    request = HttpRequest("GET", TEST_URL + original_url)
    pipeline_request = PipelineRequest(request, None)
    query_param_policy = QueryParamPolicy()

    class MockNext:
        def __init__(self):
            self.captured_url = None

        def send(self, request):
            self.captured_url = request.http_request.url
            return None

    mock_next = MockNext()
    query_param_policy.next = mock_next
    query_param_policy.send(pipeline_request)
    assert mock_next.captured_url == TEST_URL + expected_url



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
from azure.appconfiguration import (
    FeatureFlagConfigurationSetting,
    FILTER_PERCENTAGE,
)
from testcase import AppConfigTestCase
from preparers import app_config_decorator
from devtools_testutils import recorded_by_proxy, set_custom_default_matcher
import json
import pytest


class TestAppConfigurationConsistency(AppConfigTestCase):
    @app_config_decorator
    @recorded_by_proxy
    def test_update_json_by_value(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        client = self.create_client(appconfiguration_connection_string)
        key = self.get_resource_name("key")
        feature_flag = FeatureFlagConfigurationSetting(
            key, enabled=True, filters=[{"name": FILTER_PERCENTAGE, "parameters": {"Value": 10, "User": "user1"}}]
        )
        set_flag = client.set_configuration_setting(feature_flag)

        set_flag.value = json.dumps(
            {
                "conditions": {
                    "client_filters": [
                        {
                            "name": "Microsoft.Targeting",
                            "parameters": {
                                "name": "Microsoft.Targeting",
                                "parameters": {"Audience": {"DefaultRolloutPercentage": 50, "Groups": [], "Users": []}},
                            },
                        }
                    ]
                },
                "description": "",
                "enabled": False,
                "id": key,
            }
        )

        set_flag = client.set_configuration_setting(set_flag)
        assert isinstance(set_flag, FeatureFlagConfigurationSetting)
        assert set_flag.enabled == False
        assert set_flag.key.endswith(key)

        client.delete_configuration_setting(set_flag)

    @app_config_decorator
    @recorded_by_proxy
    def test_feature_flag_invalid_json(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        client = self.create_client(appconfiguration_connection_string)
        key = self.get_resource_name("key")
        feature_flag = FeatureFlagConfigurationSetting(key, enabled=True)
        set_flag = client.set_configuration_setting(feature_flag)

        with pytest.raises(TypeError):
            set_flag.value = []
            client.set_configuration_setting(set_flag)

        client.delete_configuration_setting(feature_flag)

    @app_config_decorator
    @recorded_by_proxy
    def test_feature_flag_invalid_json_string(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        client = self.create_client(appconfiguration_connection_string)
        key = self.get_resource_name("key")
        feature_flag = FeatureFlagConfigurationSetting(key, enabled=True)
        set_flag = client.set_configuration_setting(feature_flag)

        set_flag.value = "hello world"
        received = client.set_configuration_setting(set_flag)

        assert isinstance(received, FeatureFlagConfigurationSetting)
        client.delete_configuration_setting(set_flag)

    @app_config_decorator
    @recorded_by_proxy
    def test_feature_flag_invalid_json_access_properties(self, appconfiguration_connection_string):
        # response header <x-ms-content-sha256> and <x-ms-date> are missing in python38.
        set_custom_default_matcher(compare_bodies=False, excluded_headers="x-ms-content-sha256,x-ms-date")
        client = self.create_client(appconfiguration_connection_string)
        key = self.get_resource_name("key")
        feature_flag = FeatureFlagConfigurationSetting(key, enabled=True)
        set_flag = client.set_configuration_setting(feature_flag)

        set_flag.value = "hello world"
        assert set_flag.enabled == False
        assert set_flag.filters == None
        client.delete_configuration_setting(feature_flag)


class TestAppConfigurationConsistencyUnitTest(AppConfigTestCase):
    def test_feature_flag_set_value(self):
        key = self.get_resource_name("key")
        feature_flag = FeatureFlagConfigurationSetting(
            key, enabled=True, filters=[{"name": FILTER_PERCENTAGE, "parameters": {"Value": 10, "User": "user1"}}]
        )
        feature_flag.value = json.dumps({"conditions": {"client_filters": []}, "enabled": False})

        assert feature_flag.enabled == False

    def test_feature_flag_set_enabled(self):
        key = self.get_resource_name("key")
        feature_flag = FeatureFlagConfigurationSetting(
            key, enabled=True, filters=[{"name": FILTER_PERCENTAGE, "parameters": {"Value": 10, "User": "user1"}}]
        )
        feature_flag.enabled = False

        temp = json.loads(feature_flag.value)
        assert temp["enabled"] == False

    def test_feature_flag_prefix(self):
        key = self.get_resource_name("key")
        feature_flag = FeatureFlagConfigurationSetting(key, enabled=True)
        assert feature_flag.key.startswith(".appconfig.featureflag/")

    def test_feature_flag_different_key_and_id(self):
        key = f'.appconfig.featureflag/{self.get_resource_name("key")}'
        feature_id = self.get_resource_name("id")
        feature_flag = FeatureFlagConfigurationSetting(feature_id, enabled=True, key=key)

        assert feature_flag.feature_id == feature_id
        assert feature_flag.key == key



# ------------------------------------
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ------------------------------------


"""Customize generated code here.

Follow our quickstart for examples: https://aka.ms/azsdk/python/dpcodegen/python/customize
"""
import json
import urllib.parse
from typing import Any, Iterable, List, Optional, Union, MutableMapping, Type
from azure.core import MatchConditions
from azure.core.exceptions import (
    ClientAuthenticationError,
    HttpResponseError,
    ResourceExistsError,
    ResourceNotFoundError,
    ResourceNotModifiedError,
    ResourceModifiedError,
    map_error,
)
from azure.core.pipeline import PipelineResponse
from azure.core.rest import HttpRequest
from azure.core.tracing.decorator import distributed_trace
from azure.core.utils import case_insensitive_dict
from ._operations import (
    AzureAppConfigurationClientOperationsMixin as AzureAppConfigClientOpGenerated,
    ClsType,
    build_azure_app_configuration_get_key_values_request,
    prep_if_match,
    prep_if_none_match,
)
from .. import models as _models
from .._model_base import _deserialize
from .._serialization import Serializer

_SERIALIZER = Serializer()
_SERIALIZER.client_side_validation = False


class AzureAppConfigurationClientOperationsMixin(AzureAppConfigClientOpGenerated):
    @distributed_trace
    def get_key_values_in_one_page(
        self,
        *,
        key: Optional[str] = None,
        label: Optional[str] = None,
        sync_token: Optional[str] = None,
        after: Optional[str] = None,
        accept_datetime: Optional[str] = None,
        select: Optional[List[Union[str, _models.ConfigurationSettingFields]]] = None,
        snapshot: Optional[str] = None,
        tags: Optional[List[str]] = None,
        etag: Optional[str] = None,
        match_condition: Optional[MatchConditions] = None,
        continuation_token: Optional[str] = None,
        **kwargs: Any
    ) -> Iterable["_models.KeyValue"]:
        """Gets a list of key-values in one page.

        Gets a list of key-values in one page.

        :keyword key: A filter used to match keys. Syntax reference:
         https://aka.ms/azconfig/docs/keyvaluefiltering. Default value is None.
        :paramtype key: str
        :keyword label: A filter used to match labels. Syntax reference:
         https://aka.ms/azconfig/docs/keyvaluefiltering. Default value is None.
        :paramtype label: str
        :keyword sync_token: Used to guarantee real-time consistency between requests. Default value is
         None.
        :paramtype sync_token: str
        :keyword after: Instructs the server to return elements that appear after the element referred
         to by the specified token. Default value is None.
        :paramtype after: str
        :keyword accept_datetime: Requests the server to respond with the state of the resource at the
         specified
         time. Default value is None.
        :paramtype accept_datetime: str
        :keyword select: Used to select what fields are present in the returned resource(s). Default
         value is None.
        :paramtype select: list[str or ~azure.appconfiguration.models.ConfigurationSettingFields]
        :keyword snapshot: A filter used get key-values for a snapshot. The value should be the name of
         the snapshot. Not valid when used with 'key' and 'label' filters. Default value is None.
        :paramtype snapshot: str
        :keyword tags: A filter used to query by tags. Syntax reference:
         https://aka.ms/azconfig/docs/keyvaluefiltering. Default value is None.
        :paramtype tags: list[str]
        :keyword etag: check if resource is changed. Set None to skip checking etag. Default value is
         None.
        :paramtype etag: str
        :keyword match_condition: The match condition to use upon the etag. Default value is None.
        :paramtype match_condition: ~azure.core.MatchConditions
        :param str continuation_token: An opaque continuation token.
        :return: An iterator like instance of KeyValue
        :rtype: ~azure.core.paging.ItemPaged[~azure.appconfiguration.models.KeyValue]
        :raises ~azure.core.exceptions.HttpResponseError:
        """
        _headers = kwargs.pop("headers", {}) or {}
        _params = kwargs.pop("params", {}) or {}

        cls: ClsType[List[_models.KeyValue]] = kwargs.pop("cls", None)

        error_map: MutableMapping[int, Type[HttpResponseError]] = {
            401: ClientAuthenticationError,
            404: ResourceNotFoundError,
            409: ResourceExistsError,
            304: ResourceNotModifiedError,
        }
        if match_condition == MatchConditions.IfNotModified:
            error_map[412] = ResourceModifiedError
        elif match_condition == MatchConditions.IfPresent:
            error_map[412] = ResourceNotFoundError
        elif match_condition == MatchConditions.IfMissing:
            error_map[412] = ResourceExistsError
        error_map.update(kwargs.pop("error_map", {}) or {})

        def prepare_request(next_link=None):
            if not next_link:

                _request = build_azure_app_configuration_get_key_values_request(
                    key=key,
                    label=label,
                    sync_token=sync_token,
                    after=after,
                    accept_datetime=accept_datetime,
                    select=select,
                    snapshot=snapshot,
                    tags=tags,
                    etag=etag,
                    match_condition=match_condition,
                    api_version=self._config.api_version,
                    headers=_headers,
                    params=_params,
                )
                path_format_arguments = {
                    "endpoint": self._serialize.url(
                        "self._config.endpoint", self._config.endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            else:
                # make call to next link with the client's api-version
                _parsed_next_link = urllib.parse.urlparse(next_link)
                _next_request_params = case_insensitive_dict(
                    {
                        key: [urllib.parse.quote(v) for v in value]
                        for key, value in urllib.parse.parse_qs(_parsed_next_link.query).items()
                    }
                )
                _next_request_params["api-version"] = self._config.api_version

                # Add etag and match_condition to headers
                _next_headers = dict(_headers)
                accept = _headers.pop("Accept", None)
                if sync_token is not None:
                    _next_headers["Sync-Token"] = _SERIALIZER.header("sync_token", sync_token, "str")
                if accept_datetime is not None:
                    _next_headers["Accept-Datetime"] = _SERIALIZER.header("accept_datetime", accept_datetime, "str")
                if accept is not None:
                    _next_headers["Accept"] = _SERIALIZER.header("accept", accept, "str")
                if_match = prep_if_match(etag, match_condition)
                if if_match is not None:
                    _next_headers["If-Match"] = _SERIALIZER.header("if_match", if_match, "str")
                if_none_match = prep_if_none_match(etag, match_condition)
                if if_none_match is not None:
                    _next_headers["If-None-Match"] = _SERIALIZER.header("if_none_match", if_none_match, "str")
                _request = HttpRequest(
                    "GET",
                    urllib.parse.urljoin(next_link, _parsed_next_link.path),
                    params=_next_request_params,
                    headers=_next_headers,
                )
                path_format_arguments = {
                    "endpoint": self._serialize.url(
                        "self._config.endpoint", self._config.endpoint, "str", skip_quote=True
                    ),
                }
                _request.url = self._client.format_url(_request.url, **path_format_arguments)

            return _request

        _request = prepare_request(continuation_token)
        _stream = False
        pipeline_response: PipelineResponse = self._client._pipeline.run(  # pylint: disable=protected-access
            _request, stream=_stream, **kwargs
        )
        response = pipeline_response.http_response

        valid_status_codes = [200]
        if etag is not None and match_condition is MatchConditions.IfModified:
            valid_status_codes.append(304)

        if response.status_code not in valid_status_codes:
            map_error(status_code=response.status_code, response=response, error_map=error_map)
            error = _deserialize(_models.Error, response.json())
            raise HttpResponseError(response=response, model=error)

        response_headers = response.headers
        deserialized = json.loads("{}")
        if response.status_code != 304:
            deserialized = pipeline_response.http_response.json()
        else:
            unparsed_link = pipeline_response.http_response.headers.get("Link")
            next_link = None
            if unparsed_link:
                next_link = unparsed_link[1 : unparsed_link.index(">")]
            deserialized["@nextLink"] = next_link

        if cls:
            return cls(pipeline_response, deserialized, response_headers)

        return deserialized


__all__: List[str] = [
    "AzureAppConfigurationClientOperationsMixin"
]  # Add all objects you want publicly available to users at this package level


def patch_sdk():
    """Do not remove from this file.

    `patch_sdk` is a last resort escape hatch that allows you to do customizations
    you can't accomplish using the techniques described in
    https://aka.ms/azsdk/python/dpcodegen/python/customize
    """
